@@ -9,16 +9,122 @@ from os import remove
 from datetime import datetime, timedelta
 
 
-class InventoryItem(Item):
-	is_favorite = False
-
+class ItemWithIsFavorite(Item):
 	def __init__(self, id: int = 0, stack: int = 0, prefix: int = 0, is_favorite: bool = False):
-		self.is_favorite = is_favorite
+		self.is_favorite: bool = is_favorite
 		super().__init__(id, stack, prefix)
 
 
+class Loadout:
+	def __init__(self):
+		self.armor_and_accessories: list[Item] = []
+		self.vanity_armor_and_accessories: list[Item] = []
+		self.dyes: list[Item] = []
+		self.is_accessories_hidden: Bitarray.bitarray = Bitarray.bitarray()
+
+
 class TerrariaPCPlayer:
+	MAX_SUPPORTED_VERSION: int = 270
+	MIN_SUPPORTED_VERSION: int = 0
+
 	def __init__(self, path: str = None, version: int = None):
+		self.version: int = self.MAX_SUPPORTED_VERSION
+		self.is_favorite: bool = False
+		self.name: str = ""
+		self.difficulty: int = 0
+		self.playtime: timedelta = timedelta()
+		self.hair_style: int = 0
+		self.hair_dye: int = 0
+		self.is_accessories_hidden: Bitarray.bitarray = Bitarray.bitarray()
+		self.is_pet_hidden: bool = False
+		self.is_light_pet_hidden: bool = False
+		self.is_male: bool = False
+		self.style: int = 0
+		self.life: int = 100
+		self.max_life: int = 100
+		self.mana: int = 20
+		self.max_mana: int = 20
+		self.has_demon_heart_accessory_slot: bool = False
+		self.is_biome_torches_unlocked: bool = False
+		self.is_using_biome_torches: bool = False
+		self.has_downed_DD2_event: bool = False
+		self.tax_money: int = 0
+		self.hair_color: Color = Color(0, 0, 0)
+		self.skin_color: Color = Color(0, 0, 0)
+		self.eye_color: Color = Color(0, 0, 0)
+		self.shirt_color: Color = Color(0, 0, 0)
+		self.under_shirt_color: Color = Color(0, 0, 0)
+		self.pants_color: Color = Color(0, 0, 0)
+		self.shoe_color: Color = Color(0, 0, 0)
+		self.armor_and_accessories: list[Item] = []
+		self.vanity_armor_and_accessories: list[Item] = []
+		self.dyes: list[Item] = []
+		self.inventory: list[list[Item | ItemWithIsFavorite]] = []
+		self.coins: list[Item] = []
+		self.ammo: list[Item] = []
+		self.misc_equips: list[Item] = []
+		self.misc_dyes: list[Item] = []
+		self.piggy_bank: list[list[Item]] = []
+		self.safe: list[list[Item]] = []
+		self.defenders_forge: list[list[Item | ItemWithIsFavorite]] = []
+		self.void_vault: list[list[Item]] = []
+		self.is_void_vault_enabled: bool = False
+		self.buffs: list[Buff] = []
+		self.spawn_points: list[SpawnPoint] = []
+		self.is_hotbar_locked: bool = False
+		self.is_time_info_hidden: bool = False
+		self.is_weather_info_hidden: bool = False
+		self.is_fishing_power_info_hidden: bool = False
+		self.is_position_info_hidden: bool = False
+		self.is_depth_info_hidden: bool = False
+		self.is_creature_count_info_hidden: bool = False
+		self.is_kill_count_info_hidden: bool = False
+		self.is_moon_phase_info_hidden: bool = False
+		self.is_movement_speed_info_hidden: bool = False
+		self.is_treasure_finder_info_hidden: bool = False
+		self.is_rare_creatures_finder_info_hidden: bool = False
+		self.is_damage_per_second_info_hidden: bool = False
+		self.angler_quests_finished: int = 0
+		self.quick_shortcuts: list[int] = [-1, -1, -1, -1]
+		self.is_ruler_enabled: bool = False
+		self.is_mechanical_ruler_enabled: bool = False
+		self.is_auto_placement_actuators_enabled: bool = False
+		self.is_auto_paint_enabled: bool = False
+		self.red_wires_visibility: int = 0
+		self.blue_wires_visibility: int = 0
+		self.green_wires_visibility: int = 0
+		self.yellow_wires_visibility: int = 0
+		self.is_showing_showing_wires_and_actuators_forced: bool = False
+		self.actuators_visibility: int = 0
+		self.is_tile_replacement_enabled: bool = True
+		self.is_biome_torches_enabled: bool = False
+		self.bartender_quest_log: int = 0
+		self.is_dead: bool = False
+		self.time_to_respawn: timedelta = timedelta()
+		self.last_time_player_was_saved: datetime = datetime.now()
+		self.golfer_score_accumulated: int = 0
+		self.item_research: dict = {}
+		self.temporary_mouse_item: Item = Item()
+		self.temporary_research_item: Item = Item()
+		self.temporary_guide_item: Item = Item()
+		self.temporary_goblin_item: Item = Item()
+		self.is_godmode_enabled: bool = False
+		self.is_far_placement_enabled: bool = True
+		self.spawn_rate: float = 0.5
+		self.is_artisan_bread_eaten: bool = False
+		self.is_aegis_crystal_used: bool = False
+		self.is_aegis_fruit_used: bool = False
+		self.is_arcane_crystal_used: bool = False
+		self.is_galaxy_pearl_used: bool = False
+		self.is_gummy_worm_used: bool = False
+		self.is_ambrosia_used: bool = False
+		self.number_of_deaths_not_caused_by_players: int = 0
+		self.number_of_deaths_caused_by_players: int = 0
+		self.is_super_cart_unlocked: bool = False
+		self.is_super_cart_enabled: bool = False
+		self.selected_loadout_index: int = 0
+		self.loadouts: list[Loadout] = []
+
 		if path is not None:
 			self.read_from_file(path, version)
 
@@ -69,92 +175,11 @@ class TerrariaPCPlayer:
 			return file.read_int32()
 
 	def reset(self):
-		self.version = 244
-		self.is_favorite = False
-		self.name = ""
-		self.difficulty = 0
-		self.playtime = timedelta()
-		self.hair_style = 0
-		self.hair_dye = 0
-		self.is_accessories_hidden = Bitarray.bitarray()
-		self.is_pet_hidden = False
-		self.is_light_pet_hidden = False
-		self.is_male = False
-		self.style = 0
-		self.life = 100
-		self.max_life = 100
-		self.mana = 20
-		self.max_mana = 20
-		self.has_demon_heart_accessory_slot = False
-		self.is_biome_torches_unlocked = False
-		self.is_using_biome_torches = False
-		self.has_downed_DD2_event = False
-		self.tax_money = 0
-		self.hair_color = Color(0, 0, 0)
-		self.skin_color = Color(0, 0, 0)
-		self.eye_color = Color(0, 0, 0)
-		self.shirt_color = Color(0, 0, 0)
-		self.under_shirt_color = Color(0, 0, 0)
-		self.pants_color = Color(0, 0, 0)
-		self.shoe_color = Color(0, 0, 0)
-		self.armor_and_accessories = []
-		self.vanity_armor_and_accessories = []
-		self.dyes = []
-		self.inventory = []
-		self.coins = []
-		self.ammo = []
-		self.misc_equips = []
-		self.misc_dyes = []
-		self.piggy_bank = []
-		self.safe = []
-		self.defenders_forge = []
-		self.void_vault = []
-		self.is_void_vault_enabled = False
-		self.buffs = []
-		self.spawn_points = []
-		self.is_hotbar_locked = False
-		self.is_time_info_hidden = False
-		self.is_weather_info_hidden = False
-		self.is_fishing_power_info_hidden = False
-		self.is_position_info_hidden = False
-		self.is_depth_info_hidden = False
-		self.is_creature_count_info_hidden = False
-		self.is_kill_count_info_hidden = False
-		self.is_moon_phase_info_hidden = False
-		self.is_movement_speed_info_hidden = False
-		self.is_treasure_finder_info_hidden = False
-		self.is_rare_creatures_finder_info_hidden = False
-		self.is_damage_per_second_info_hidden = False
-		self.angler_quests_finished = 0
-		self.quick_shortcuts = [-1, -1, -1, -1]
-		self.is_ruler_enabled = False
-		self.is_mechanical_ruler_enabled = False
-		self.is_auto_placement_actuators_enabled = False
-		self.is_auto_paint_enabled = False
-		self.red_wires_visibility = 0
-		self.blue_wires_visibility = 0
-		self.green_wires_visibility = 0
-		self.yellow_wires_visibility = 0
-		self.is_showing_showing_wires_and_actuators_forced = False
-		self.actuators_visibility = 0
-		self.is_tile_replacement_enabled = True
-		self.is_biome_torches_enabled = False
-		self.bartender_quest_log = 0
-		self.is_dead = False
-		self.time_to_respawn = timedelta()
-		self.last_time_player_was_saved = datetime(1, 1, 1)
-		self.golfer_score_accumulated = 0
-		self.item_research = {}
-		self.temporary_mouse_item = Item()
-		self.temporary_research_item = Item()
-		self.temporary_guide_item = Item()
-		self.temporary_goblin_item = Item()
-		self.is_godmode_enabled = False
-		self.is_far_placement_enabled = True
-		self.spawn_rate = 0.5
+		self.__init__()
 
 	def read_from_file(self, path: str, version: int = None) -> None:
 		with TerrariaPCPlayer.BinaryReaderDecrypt(path, bytes.fromhex("6800330079005F006700550079005A00")) as file:
+			self.reset()
 			self.version = file.read_int32()
 			if version is None:
 				version = self.version
@@ -164,9 +189,9 @@ class TerrariaPCPlayer:
 			def read_color() -> Color:
 				return Color(file.read_byte(), file.read_byte(), file.read_byte())
 
-			def read_item(disable_stack: bool = False, disable_prefix: bool = False) -> Item:  # normal read item
+			def read_item(*, disable_stack: bool = False, disable_prefix: bool = False) -> Item:  # normal read item
 				if version < 38:
-					return read_item_legacy(disable_stack, disable_prefix)
+					return read_item_legacy(disable_stack=disable_stack, disable_prefix=disable_prefix)
 				id = file.read_int32()
 				stack = 1
 				prefix = 0
@@ -187,7 +212,7 @@ class TerrariaPCPlayer:
 
 				return Item(id, stack, prefix)
 
-			def read_item_legacy(disable_stack: bool = False, disable_prefix: bool = False) -> Item:
+			def read_item_legacy(*, disable_stack: bool = False, disable_prefix: bool = False) -> Item:
 				name = file.read_string()
 				stack = 1
 				prefix = 0
@@ -237,17 +262,22 @@ class TerrariaPCPlayer:
 						bitarray.append(byte >> x & 1)
 				return bitarray
 
-			def read_container(size_x: int = 10, size_y: int = 4):
+			def read_container(size_x: int = 10, size_y: int = 4, *, read_is_favorite_data: bool = False):
 				if not isinstance(size_x, int):
 					raise TypeError("size_x must be an int")
 				if not isinstance(size_y, int):
 					raise TypeError("size_y must be an int")
+				if not isinstance(read_is_favorite_data, bool):
+					raise TypeError("read_is_favorite_data must be a bool")
 
 				container = []
 				for _ in range(0, size_y):
 					row = []
 					for _ in range(0, size_x):
-						row.append(read_item())
+						item = read_item()
+						if read_is_favorite_data:
+							item = ItemWithIsFavorite(item.id, item.stack, item.prefix, file.read_bool())
+						row.append(item)
 					container.append(row)
 				return container
 
@@ -316,12 +346,29 @@ class TerrariaPCPlayer:
 				self.is_biome_torches_unlocked = file.read_bool()
 				self.is_using_biome_torches = file.read_bool()
 
+			if version >= 256:
+				self.is_artisan_bread_eaten = file.read_bool()
+
+			if version >= 260:
+				self.is_aegis_crystal_used = file.read_bool()
+				self.is_aegis_fruit_used = file.read_bool()
+				self.is_arcane_crystal_used = file.read_bool()
+				self.is_galaxy_pearl_used = file.read_bool()
+				self.is_gummy_worm_used = file.read_bool()
+				self.is_ambrosia_used = file.read_bool()
+
 			if version >= 182:
 				self.has_downed_DD2_event = file.read_bool()
 
 			if version >= 128:
 				self.tax_money = file.read_int32()
 
+			if version >= 254:
+				self.number_of_deaths_not_caused_by_players = file.read_int32()
+
+			if version >= 254:
+				self.number_of_deaths_caused_by_players = file.read_int32()
+
 			self.hair_color = read_color()
 			self.skin_color = read_color()
 			self.eye_color = read_color()
@@ -360,21 +407,21 @@ class TerrariaPCPlayer:
 				for _ in range(0, 10):
 					item = read_item()
 					if version >= 114:
-						item = InventoryItem(item.id, item.stack, item.prefix, file.read_bool())
+						item = ItemWithIsFavorite(item.id, item.stack, item.prefix, file.read_bool())
 					row.append(item)
 				self.inventory.append(row)
 
 			for _ in range(0, 4):
 				item = read_item()
 				if version >= 114:
-					item = InventoryItem(item.id, item.stack, item.prefix, file.read_bool())
+					item = ItemWithIsFavorite(item.id, item.stack, item.prefix, file.read_bool())
 				self.coins.append(item)
 
 			if version >= 15:
 				for _ in range(0, 4):
 					item = read_item()
 					if version >= 114:
-						item = InventoryItem(item.id, item.stack, item.prefix, file.read_bool())
+						item = ItemWithIsFavorite(item.id, item.stack, item.prefix, file.read_bool())
 					self.ammo.append(item)
 
 			if version >= 117:
@@ -398,13 +445,16 @@ class TerrariaPCPlayer:
 				self.defenders_forge = read_container()
 
 			if version >= 198:
-				self.void_vault = read_container()
+				self.void_vault = read_container(read_is_favorite_data=(version >= 255))
 
 			if version >= 199:
 				self.is_void_vault_enabled = file.read_bool()
 
 			if version >= 11:
-				amount = 22
+				amount = 44
+				if version < 252:
+					amount = 22
+
 				if version < 74:
 					amount = 10
 
@@ -474,7 +524,7 @@ class TerrariaPCPlayer:
 					self.time_to_respawn = timedelta(microseconds=int(file.read_int32() * 16666.66))
 
 			if version >= 202:
-				self.last_time_player_was_saved += timedelta(microseconds=int((file.read_uint64() & 4611686018427387903) / 10))
+				self.last_time_player_was_saved = datetime(1, 1, 1) + timedelta(microseconds=int((file.read_uint64() & 4611686018427387903) / 10))
 
 			if version >= 206:
 				self.golfer_score_accumulated = file.read_int32()
@@ -505,6 +555,29 @@ class TerrariaPCPlayer:
 					elif key == 14:
 						self.spawn_rate = file.read_single()
 
+			if version >= 253:
+				byte = file.read_byte()
+				self.is_super_cart_unlocked = bool(byte & 1)
+				self.is_super_cart_enabled = bool(byte & 2)
+
+			if version >= 262:
+				self.selected_loadout_index = file.read_int32()
+				for _ in range(0, 3):
+					loadout = Loadout()
+					for x in range(0, 10):
+						loadout.armor_and_accessories.append(read_item())
+
+					for x in range(0, 10):
+						loadout.vanity_armor_and_accessories.append(read_item())
+
+					for x in range(0, 10):
+						loadout.dyes.append(read_item())
+
+					for x in range(0, 10):
+						loadout.is_accessories_hidden.append(file.read_bool())
+
+					self.loadouts.append(loadout)
+
 	def write_to_file(self, path: str, version: int = None) -> None:
 		if version is None:
 			version = self.version
@@ -606,7 +679,7 @@ class TerrariaPCPlayer:
 						byte += 2 ** x2 * try_read_index(bitarray, x1 * 8 + x2, 0)
 					file.write_byte(byte)
 
-			def write_container(container: list = None, size_x: int = 10, size_y: int = 4):
+			def write_container(container: list[list[Item | ItemWithIsFavorite]] = None, size_x: int = 10, size_y: int = 4, *, write_is_favorite_data: bool = False):
 				if container is None:
 					container = []
 				if not isinstance(container, list):
@@ -620,7 +693,13 @@ class TerrariaPCPlayer:
 					if not isinstance(try_read_index(container, x1, []), list):
 						raise TypeError("Elements of container must be a list")
 					for x2 in range(0, size_x):
-						write_item(try_read_index(try_read_index(container, x1, []), x2, Item()))
+						item = try_read_index(try_read_index(container, x1, []), x2, Item())
+						write_item(item)
+						if write_is_favorite_data:
+							try:
+								file.write_bool(item.is_favorite)
+							except AttributeError:
+								file.write_bool(False)
 
 			file.write_int32(version)
 
@@ -696,12 +775,29 @@ class TerrariaPCPlayer:
 				file.write_bool(self.is_biome_torches_unlocked)
 				file.write_bool(self.is_using_biome_torches)
 
+			if version >= 256:
+				file.write_bool(self.is_artisan_bread_eaten)
+
+			if version >= 260:
+				file.write_bool(self.is_aegis_crystal_used)
+				file.write_bool(self.is_aegis_fruit_used)
+				file.write_bool(self.is_arcane_crystal_used)
+				file.write_bool(self.is_galaxy_pearl_used)
+				file.write_bool(self.is_gummy_worm_used)
+				file.write_bool(self.is_ambrosia_used)
+
 			if version >= 182:
 				file.write_bool(self.has_downed_DD2_event)
 
 			if version >= 128:
 				file.write_int32(self.tax_money)
 
+			if version >= 254:
+				file.write_int32(self.number_of_deaths_not_caused_by_players)
+
+			if version >= 254:
+				file.write_int32(self.number_of_deaths_caused_by_players)
+
 			write_color(self.hair_color)
 			write_color(self.skin_color)
 			write_color(self.eye_color)
@@ -741,24 +837,12 @@ class TerrariaPCPlayer:
 			if version >= 58:
 				amount = 5
 
-			if not isinstance(self.inventory, list):
-				raise TypeError("inventory must be an list")
-			for x1 in range(0, amount):
-				if not isinstance(try_read_index(self.inventory, x1, []), list):
-					raise TypeError("Elements of inventory must be a list")
-				for x2 in range(0, 10):
-					item = try_read_index(try_read_index(self.inventory, x1, []), x2, InventoryItem())
-					write_item(item)
-					if version >= 114:
-						try:
-							file.write_bool(item.is_favorite)
-						except AttributeError:
-							file.write_bool(False)
+			write_container(self.inventory, 10, amount, write_is_favorite_data=(version >= 114))
 
 			if not isinstance(self.coins, list):
 				raise TypeError("coins must be an list")
 			for x in range(0, 4):
-				item = try_read_index(self.coins, x, InventoryItem())
+				item = try_read_index(self.coins, x, ItemWithIsFavorite())
 				write_item(item)
 				if version >= 114:
 					try:
@@ -770,7 +854,7 @@ class TerrariaPCPlayer:
 				if not isinstance(self.ammo, list):
 					raise TypeError("ammo must be an list")
 				for x in range(0, 4):
-					item = try_read_index(self.ammo, x, InventoryItem())
+					item = try_read_index(self.ammo, x, ItemWithIsFavorite())
 					write_item(item)
 					if version >= 114:
 						try:
@@ -799,13 +883,16 @@ class TerrariaPCPlayer:
 				write_container(self.defenders_forge)
 
 			if version >= 198:
-				write_container(self.void_vault)
+				write_container(self.void_vault, write_is_favorite_data=(version >= 255))
 
 			if version >= 199:
 				file.write_bool(self.is_void_vault_enabled)
 
 			if version >= 11:
-				amount = 22
+				amount = 44
+				if version < 252:
+					amount = 22
+
 				if version < 74:
 					amount = 10
 
@@ -950,89 +1037,45 @@ class TerrariaPCPlayer:
 
 				file.write_bool(False)
 
-	MAX_SUPPORTED_VERSION = 244
-	MIN_SUPPORTED_VERSION = 0
+			if version >= 253:
+				if self.is_super_cart_unlocked != 0 and self.is_super_cart_unlocked != 1:
+					raise ValueError(f"is_super_cart_unlocked must be true or false")
+				if self.is_super_cart_enabled != 0 and self.is_super_cart_enabled != 1:
+					raise ValueError(f"is_super_cart_enabled must be true or false")
+				byte = self.is_super_cart_unlocked
+				byte |= self.is_super_cart_enabled << 1
+				file.write_byte(byte)
 
-	version = 244
-	is_favorite = False
-	name = ""
-	difficulty = 0
-	playtime = timedelta()
-	hair_style = 0
-	hair_dye = 0
-	is_accessories_hidden = Bitarray.bitarray()
-	is_pet_hidden = False
-	is_light_pet_hidden = False
-	is_male = False
-	style = 0
-	life = 100
-	max_life = 100
-	mana = 20
-	max_mana = 20
-	has_demon_heart_accessory_slot = False
-	is_biome_torches_unlocked = False
-	is_using_biome_torches = False
-	has_downed_DD2_event = False
-	tax_money = 0
-	hair_color = Color(0, 0, 0)
-	skin_color = Color(0, 0, 0)
-	eye_color = Color(0, 0, 0)
-	shirt_color = Color(0, 0, 0)
-	under_shirt_color = Color(0, 0, 0)
-	pants_color = Color(0, 0, 0)
-	shoe_color = Color(0, 0, 0)
-	armor_and_accessories = []
-	vanity_armor_and_accessories = []
-	dyes = []
-	inventory = []
-	coins = []
-	ammo = []
-	misc_equips = []
-	misc_dyes = []
-	piggy_bank = []
-	safe = []
-	defenders_forge = []
-	void_vault = []
-	is_void_vault_enabled = False
-	buffs = []
-	spawn_points = []
-	is_hotbar_locked = False
-	is_time_info_hidden = False
-	is_weather_info_hidden = False
-	is_fishing_power_info_hidden = False
-	is_position_info_hidden = False
-	is_depth_info_hidden = False
-	is_creature_count_info_hidden = False
-	is_kill_count_info_hidden = False
-	is_moon_phase_info_hidden = False
-	is_movement_speed_info_hidden = False
-	is_treasure_finder_info_hidden = False
-	is_rare_creatures_finder_info_hidden = False
-	is_damage_per_second_info_hidden = False
-	angler_quests_finished = 0
-	quick_shortcuts = [-1, -1, -1, -1]
-	is_ruler_enabled = False
-	is_mechanical_ruler_enabled = False
-	is_auto_placement_actuators_enabled = False
-	is_auto_paint_enabled = False
-	red_wires_visibility = 0
-	blue_wires_visibility = 0
-	green_wires_visibility = 0
-	yellow_wires_visibility = 0
-	is_showing_showing_wires_and_actuators_forced = False
-	actuators_visibility = 0
-	is_tile_replacement_enabled = False
-	is_biome_torches_enabled = False
-	bartender_quest_log = 0
-	is_dead = False
-	time_to_respawn = timedelta()
-	last_time_player_was_saved = datetime(1, 1, 1)
-	golfer_score_accumulated = 0
-	item_research = {}  # key is item id, value is how many items are given to research
-	temporary_mouse_item = Item()
-	temporary_research_item = Item()
-	temporary_guide_item = Item()
-	temporary_goblin_item = Item()
-	is_godmode_enabled = False
-	is_far_placement_enabled = True
-	spawn_rate = 0.5
+			if version >= 262:
+				file.write_int32(self.selected_loadout_index)
+				if not isinstance(self.loadouts, list):
+					raise TypeError("loadouts must be an list")
+
+				for x1 in range(0, 3):
+					loadout = try_read_index(self.loadouts, x1, Loadout())
+					if not isinstance(loadout, Loadout):
+						raise TypeError("Elements of loadouts must be an Loadout")
+
+					if not isinstance(loadout.armor_and_accessories, list):
+						raise TypeError("armor_and_accessories in loadout must be an list")
+
+					for x in range(0, 10):
+						write_item(try_read_index(loadout.armor_and_accessories, x, Item()))
+
+					if not isinstance(loadout.vanity_armor_and_accessories, list):
+						raise TypeError("vanity_armor_and_accessories in loadout must be an list")
+
+					for x in range(0, 10):
+						write_item(try_read_index(loadout.vanity_armor_and_accessories, x, Item()))
+
+					if not isinstance(loadout.dyes, list):
+						raise TypeError("dyes in loadout must be an list")
+
+					for x in range(0, 10):
+						write_item(try_read_index(loadout.dyes, x, Item()))
+
+					if not isinstance(loadout.is_accessories_hidden, Bitarray.bitarray):
+						raise TypeError("is_accessories_hidden in loadout must be an Bitarray")
+
+					for x in range(0, 10):
+						file.write_bool(try_read_index(loadout.is_accessories_hidden, x, False))
